#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "microwine.h"

char *mybase;
void *IMAGE_BASE;
void (*entry_point)(void) = NULL;

#define EXE_SIZE 4328960

void thunk()
{
	printf("Error: Unknown WinAPI called\n");
	exit(1);
}

extern void GetProcessHeap();
extern void GetVersionExA();
extern void HeapCreate();
extern void HeapFree();
extern void HeapAlloc();
extern void GetProcAddress();
extern void GetModuleHandleA();
extern void HeapSetInformation();
extern void FlsGetValue();
extern void FlsSetValue();
extern void FlsAlloc();
extern void EnterCriticalSection();
extern void LeaveCriticalSection();
extern void InitializeCriticalSection();
extern void GetStartupInfoA();
extern void GetSystemTimeAsFileTime();
extern void GetCurrentProcessId();
extern void GetCurrentThreadId();
extern void GetTickCount();
extern void QueryPerformanceCounter();
extern void GetStdHandle();
extern void GetFileType();
extern void SetHandleCount();
extern void GetCommandLineA();
extern void GetEnvironmentStringsW();
extern void WideCharToMultiByte();
extern void FreeEnvironmentStringsW();
extern void GetEnvironmentStrings();
extern void FreeEnvironmentStringsA();
extern void GetLastError();
extern void SetLastError();
extern void GetACP();
extern void GetCPInfo();
extern void GetModuleFileNameA();
extern void SetUnhandledExceptionFilter();
extern void HeapSize();
extern void CreateFileA();
extern void ReadFile();
extern void WriteFile();
extern void GetSystemInfo();
extern void CreateFileMappingA();
extern void OpenFileMappingA();
extern void MapViewOfFile();
extern void UnmapViewOfFile();
extern void CreatePipe();
extern void GetCurrentProcess();
extern void DuplicateHandle();
extern void CloseHandle();
extern void CreateProcessA();
extern void Sleep();
extern void GetConsoleMode();
extern void SetConsoleMode();
extern void FlushConsoleInputBuffer();
extern void GetNumberOfConsoleInputEvents();
extern void SetPriorityClass();
extern void GetLogicalDrives();
extern void SetFilePointer();
extern void FlushFileBuffers();
extern void ExitProcess();

char *argv0;
extern char cmdline[512];
extern unsigned num_cpus;

#define PATCH_FUNC(x) else if (strcmp(name, #x) == 0) { /* printf("patching 0x%x from 0x%x to %s\n", offset, *ptr, #x); */ *ptr = (long)x; }

void fixup_code(char *base, unsigned length)
{
	// 65 4c 8b 1c 25 10 00     mov    %gs:0x10,%r11
	unsigned char *ptr = (unsigned char *)memmem(base, length, "\x65\x4c\x8b\x1c\x25\x10\x00\x00\x00", 9);
	if (ptr != NULL) {
		ptr[0] = ptr[1] = ptr[2] = ptr[3] = ptr[4] = ptr[5] = ptr[6] = ptr[7] = ptr[8] = 0x90;   // nop
	}
}

void patch_iat(char *name, long offset)
{
	long *ptr = (long *)(mybase + offset);
	if (0) {
	}
	PATCH_FUNC(HeapFree)
	PATCH_FUNC(HeapAlloc)
	PATCH_FUNC(GetProcAddress)
	PATCH_FUNC(GetModuleHandleA)
	PATCH_FUNC(EnterCriticalSection)
	PATCH_FUNC(LeaveCriticalSection)
	PATCH_FUNC(GetVersionExA)
	PATCH_FUNC(GetProcessHeap)
	PATCH_FUNC(HeapSetInformation)
	PATCH_FUNC(HeapCreate)
	PATCH_FUNC(FlsGetValue)
	PATCH_FUNC(FlsSetValue)
	PATCH_FUNC(FlsAlloc)
	PATCH_FUNC(GetStartupInfoA)
	PATCH_FUNC(InitializeCriticalSection)
	PATCH_FUNC(GetSystemTimeAsFileTime)
	PATCH_FUNC(GetCurrentProcessId)
	PATCH_FUNC(GetCurrentThreadId)
	PATCH_FUNC(GetTickCount)
	PATCH_FUNC(QueryPerformanceCounter)
	PATCH_FUNC(GetStdHandle)
	PATCH_FUNC(GetFileType)
	PATCH_FUNC(SetHandleCount)
	PATCH_FUNC(GetCommandLineA)
	PATCH_FUNC(GetEnvironmentStringsW)
	PATCH_FUNC(WideCharToMultiByte)
	PATCH_FUNC(FreeEnvironmentStringsW)
	PATCH_FUNC(GetEnvironmentStrings)
	PATCH_FUNC(FreeEnvironmentStringsA)
	PATCH_FUNC(GetLastError)
	PATCH_FUNC(SetLastError)
	PATCH_FUNC(GetACP)
	PATCH_FUNC(GetCPInfo)
	PATCH_FUNC(GetModuleFileNameA)
	PATCH_FUNC(SetUnhandledExceptionFilter)
	PATCH_FUNC(HeapSize)
	PATCH_FUNC(CreateFileA)
	PATCH_FUNC(ReadFile)
	PATCH_FUNC(WriteFile)
	PATCH_FUNC(GetSystemInfo)
	PATCH_FUNC(CreateFileMappingA)
	PATCH_FUNC(OpenFileMappingA)
	PATCH_FUNC(MapViewOfFile)
	PATCH_FUNC(UnmapViewOfFile)
	PATCH_FUNC(CreatePipe)
	PATCH_FUNC(GetCurrentProcess)
	PATCH_FUNC(DuplicateHandle)
	PATCH_FUNC(CloseHandle)
	PATCH_FUNC(CreateProcessA)
	PATCH_FUNC(Sleep)
	PATCH_FUNC(GetConsoleMode)
	PATCH_FUNC(SetConsoleMode)
	PATCH_FUNC(FlushConsoleInputBuffer)
	PATCH_FUNC(GetNumberOfConsoleInputEvents)
	PATCH_FUNC(SetPriorityClass)
	PATCH_FUNC(GetLogicalDrives)
	PATCH_FUNC(SetFilePointer)
	PATCH_FUNC(FlushFileBuffers)
	PATCH_FUNC(ExitProcess)
	else {
		//printf("Warning: No patch for '%s' [offset=%x]\n", name, offset);
		//*ptr = (long)thunk;
	}
}

void decode_header(unsigned char *base)
{
	IMAGE_DOS_HEADER *dos = (IMAGE_DOS_HEADER *)base;
	IMAGE_NT_HEADERS64 *nthdr = (IMAGE_NT_HEADERS64 *)(base + dos->e_lfanew);
	IMAGE_SECTION_HEADER *sections;
	IMAGE_DATA_DIRECTORY *dir_imp, *dir_iat;
	int i;

	if (nthdr->FileHeader.Machine != 0x8664) {
		printf("no amd64 executable\n");
		exit(1);
	}

	//printf("Expected image base: 0x%x\n", nthdr->OptionalHeader.ImageBase);
	//printf("Relative address of entry point: 0x%x\n", nthdr->OptionalHeader.AddressOfEntryPoint);

	IMAGE_BASE = (void *)nthdr->OptionalHeader.ImageBase;
	sections = (IMAGE_SECTION_HEADER *)((char *)(&nthdr->OptionalHeader) + nthdr->FileHeader.SizeOfOptionalHeader);
	//printf("First section in RAM at %p\n", sections);

	// "map" in all sections
	for (i = 0; i < nthdr->FileHeader.NumberOfSections; ++i) {
#if 0
		printf("Mapping up section: %s (address=0x%x, real_address=0x%x, fileptr=0x%x, size=0x%x, relocs=%u)\n", sections[i].Name,
			nthdr->OptionalHeader.ImageBase + sections[i].VirtualAddress,
			mybase + sections[i].VirtualAddress,
			sections[i].PointerToRawData, sections[i].SizeOfRawData,
			sections[i].NumberOfRelocations);
#endif
		memcpy(mybase + sections[i].VirtualAddress, base + sections[i].PointerToRawData,
			sections[i].SizeOfRawData);

		if (strcmp((char *)sections[i].Name, ".text") == 0) {
			fixup_code(mybase + sections[i].VirtualAddress, sections[i].SizeOfRawData);
		}
	}

	// fix up some imports
	dir_imp = &nthdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT];
	dir_iat = &nthdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IAT];

#if 0
	printf("imports at 0x%x, iat at 0x%x\n",
		nthdr->OptionalHeader.ImageBase + dir_imp->VirtualAddress,
		nthdr->OptionalHeader.ImageBase + dir_iat->VirtualAddress);
#endif

	for (i = 0; i < dir_imp->Size / sizeof(IMAGE_IMPORT_DESCRIPTOR); ++i) {
		IMAGE_IMPORT_DESCRIPTOR *import =
			(IMAGE_IMPORT_DESCRIPTOR *)(mybase + dir_imp->VirtualAddress + i * sizeof(IMAGE_IMPORT_DESCRIPTOR));
		IMAGE_THUNK_DATA64 *thunk;
		//long *iat = (long *)(mybase + import->FirstThunk);
		long iat = import->FirstThunk;

		if (import->u.Characteristics == 0)
			break;

		thunk = (IMAGE_THUNK_DATA64 *)(mybase + import->u.OriginalFirstThunk);
		/* printf("import dll name 0x%x (%s), rva to thunk=0x%x\n", import->Name, mybase + import->Name,
			import->FirstThunk); */

		while (thunk->u1.Ordinal) {
			IMAGE_IMPORT_BY_NAME *iibn = (IMAGE_IMPORT_BY_NAME *)(mybase + thunk->u1.AddressOfData);
//			printf("  import func name %s, iat entry = 0x%x\n", iibn->Name, *iat);

			//patch_iat(iibn->Name, *iat);
			patch_iat((char *)iibn->Name, iat);

			++thunk;
			iat += 8;
		}
	}

	// compute the entry point
	entry_point = (void (*)(void))(mybase + nthdr->OptionalHeader.AddressOfEntryPoint);
}

unsigned get_num_cpus(void)
{
	char buf[512];
	unsigned num_cpus = 0;

	FILE *cpuinfo = fopen("/proc/cpuinfo", "r");
	if (cpuinfo == NULL) {
		perror("/proc/cpuinfo");
		exit(1);
	}

	while (!feof(cpuinfo)) {
		if (fgets(buf, 512, cpuinfo) == NULL)
			break;

		if (strncmp(buf, "processor\t:", 11) == 0) {
			++num_cpus;
		}
	}

	fclose(cpuinfo);
	return num_cpus;
}
	
void build_cmdline(int argc, char **argv)
{
	int i;
	strcpy(cmdline, argv[1]);

	for (i = 2; i < argc; ++i) {
		strcat(cmdline, " ");
		strcat(cmdline, argv[i]);
	}
}

int main(int argc, char **argv)
{
	int fd;
	unsigned char *base;

	if (argc == 1) {
		fprintf(stderr, "Usage: ./microwine RYBKA-EXECUTABLE [arguments]\n");
		exit(1);
	}

	fd = open(argv[1], O_RDONLY);
	if (fd == -1) {
		perror(argv[1]);
		exit(1);
	}

	// allocate 32MB, in which we will "map" up everything
	//mybase = (unsigned char *)malloc(32 * 1048576);
	mybase = mmap((void *)0x400000, 32 * 1048576, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_FIXED|MAP_ANONYMOUS, -1, 0);
	if (mybase == NULL) {
		perror("malloc");
		exit(1);
	}
	
	// initialize some variables we'll need
	argv0 = argv[0];
	num_cpus = get_num_cpus();
	build_cmdline(argc, argv);

	//base = mmap(IMAGE_BASE + 0x100000000, EXE_SIZE, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_FIXED, fd, 0);
	base = mmap(NULL, EXE_SIZE, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE, fd, 0);
	if (base == MAP_FAILED) {
		perror("mmap");
		exit(1);
	}

	decode_header(base);

	//entry_point = (mybase + ENTRYPOINT_RVA);
	entry_point();
	return 0;
}

unsigned CreateProcessA_cthunk(char *cmdline, LPSTARTUPINFO si, LPPROCESS_INFORMATION pi)
{
	char *cl;
	char *args[32];
	int i;
	pid_t pid;

	pid = fork();
	
	switch (pid) {
	case -1:
		perror("fork()");
		exit(1);
	case 0: /* child */
		break;
	default: /* parent */
		pi->hProcess = pi->hThread = (HANDLE)(long)pid;
		pi->dwProcessId = pi->dwThreadId = pid;
		return 1;
	}

	if (si != NULL && ((si->dwFlags & STARTF_USESTDHANDLES) != 0)) {
		if (si->hStdInput != NULL) {
			dup2((long)si->hStdInput, 0);
		}
		if (si->hStdOutput != NULL) {
			dup2((long)si->hStdInput, 1);
		}
		if (si->hStdError != NULL) {
			dup2((long)si->hStdInput, 2);
		}
	}

	// split the command line
	cl = strdup(cmdline);
	args[0] = argv0;
	args[1] = strdup(strtok(cl, " "));
	
	i = 2;
	for ( ;; ) {
		char *ptr = strtok(NULL, " ");

		if (ptr == NULL) {
			args[i] = NULL;
			break;
		}

		args[i] = strdup(ptr);

		++i;
	}

	free(cl);
	
	if (execvp(argv0, args) == -1) {
		perror(argv0);
		exit(1);
	}
	return 0;
}
